﻿using System.Collections.Generic;
using System.Linq;

namespace MyMapper
{
    /// <summary>
    /// Base class for mapper implementations. Defines methods that all mappers must have.
    /// </summary>
    public abstract class MapperBase
    {
        /// <summary>
        /// Maps the given types using given property map. This method can be used to provide 
        /// mapper with custom property map.
        /// </summary>
        /// <typeparam name="T">Source object where property values are copied from.</typeparam>
        /// <typeparam name="TU">Target object where property values are copied to.</typeparam>
        /// <param name="maps">Custom property map for given two types.</param>
        public abstract void MapTypes<T, TU>(IList<PropertyMap> maps);

        /// <summary>
        /// Creates the mapping between given types. This method uses base class property 
        /// mapper to get property mapping.
        /// </summary>
        /// <typeparam name="T">Source type from where property values are copied.</typeparam>
        /// <typeparam name="TU">Target type to where property values are copied.</typeparam>
        public abstract void MapTypes<T, TU>();

        /// <summary>
        /// Copies property values from source object to target object using previously created 
        /// mapping. Mapping must exist before copy operation is called.
        /// </summary>
        /// <typeparam name="T">Type of source object.</typeparam>
        /// <typeparam name="TU">Type of target object.</typeparam>
        /// <param name="source">Source object where property values are copied from.</param>
        /// <param name="target">Target object where property values are copied to.</param>
        public abstract void Copy<T, TU>(T source, TU target);

        /// <summary>
        /// Created property map for given source and target type.
        /// </summary>
        /// <typeparam name="T">Source type of mapping.</typeparam>
        /// <typeparam name="TU">Target type of mapping.</typeparam>
        /// <returns>List of mapped properties between given types.</returns>
        protected virtual IList<PropertyMap> GetMatchingProperties<T, TU>()
        {
            var sourceProperties = typeof(T).GetProperties();
            var targetProperties = typeof(TU).GetProperties();

            var properties = (from s in sourceProperties
                              from t in targetProperties
                              where s.Name == t.Name &&
                                    s.CanRead &&
                                    t.CanWrite &&
                                    s.PropertyType == t.PropertyType &&
                                    (
                                        (
                                            s.PropertyType.IsValueType &&
                                            t.PropertyType.IsValueType
                                        ) ||
                                        (
                                            s.PropertyType == typeof(string) &&
                                            t.PropertyType == typeof(string)
                                        )
                                    )
                              select new PropertyMap
                              {
                                  SourceProperty = s,
                                  TargetProperty = t
                              }).ToList();
            return properties;
        }

        /// <summary>
        /// Generated unique key for given types.
        /// </summary>
        /// <typeparam name="T">Type of source object.</typeparam>
        /// <typeparam name="TU">Type of target object.</typeparam>
        /// <returns>String key that is unique for given source and target type.</returns>
        protected virtual string GetMapKey<T, TU>()
        {
            var className = "Copy_";
            className += typeof(T).FullName.Replace(".", "_");
            className += "_";
            className += typeof(TU).FullName.Replace(".", "_");

            return className;
        }
    }
}